from datetime import datetime
from typing import Optional
from telegram import Update
from ..trigger_result import TriggerResult
from .basic_triggers import BaseTrigger


class EventStreakTrigger(BaseTrigger):
    def __init__(self,
                 streak_timeout_sec: int,
                 event_count: int,
                 counting_event_trigger: BaseTrigger,
                 resetting_event_trigger: Optional[BaseTrigger] = None):
        self.streak_timeout_sec = streak_timeout_sec
        self.event_count = event_count
        self.counting_event_trigger = counting_event_trigger

        if resetting_event_trigger is not None:
            self.resetting_event_trigger = resetting_event_trigger
        else:
            self.resetting_event_trigger = ~counting_event_trigger

        self.message_counter = 0
        self.streak_start_timestamp = 0

    def check_trigger(self, update: Update) -> TriggerResult:
        if self.counting_event_trigger.check_trigger(update).should_respond:
            now_timestamp = int(datetime.now().timestamp())

            if now_timestamp - self.streak_start_timestamp > self.streak_timeout_sec:
                self.reset_count()

            if self.message_counter == 0:
                self.streak_start_timestamp = int(datetime.now().timestamp())

            self.message_counter += 1

            if self.message_counter == self.event_count:
                self.reset_count()
                return TriggerResult(should_respond=True)

        elif self.resetting_event_trigger.check_trigger(update).should_respond:
            self.reset_count()

        return TriggerResult(should_respond=False)

    def reset_count(self):
        self.message_counter = 0


import json
from pathlib import Path
from gramhopper.configuration import configuration_dir
from gramhopper.handlers.handler import Handler
from gramhopper.responses import Responses
from gramhopper.triggers import Triggers


with open(Path(configuration_dir(), 'users.json'), 'r') as users_file:
    users = json.load(users_file)

hearts = ['❤', '💛', '💚', '💙', '💜', '🖤']
COOP_GIF_URL = 'https://media.giphy.com/media/3o6Zt9bsidO7reLQPe/giphy.gif'
palti_reactable_adjectives = ['יפה', 'מושלם', 'נהדר']
palti_you_are_regex = '^(?:' + 'זה ' + ')?(' + '|'.join(palti_reactable_adjectives) + ')$'
whorehouses_inputs = ['ביטוח לאומי', 'שלישות', 'משא"ן', 'בית זונות', 'דואר ישראל', 'רכבת ישראל', 'הגנ"ק', 'משו"ב', "מחנה ידין"]
# Reaction is sampled uniformly from this list. Repetition is used to change each reaction's probability.
whorehouses_replies = (['1500 שקל קנס'] * 4) + ['#להפריט']

rule_handlers = [
    Handler(Triggers.text.has_exact_word('סתום'), Responses.preset.reply('גום')),
    Handler(Triggers.text.has_exact_word('גום'), Responses.preset.reply('סתום')),
    Handler(Triggers.text.has_exact_word(['Yafe', 'yafe']), Responses.preset.reply('רמאי מסריח')),
    Handler(Triggers.text.has_substring(hearts), Responses.preset.message(hearts)),
    Handler(Triggers.text.has_substring(whorehouses_inputs), Responses.preset.reply(whorehouses_replies)),
    Handler(Triggers.text.regexp('^(לול)$'), Responses.preset.document(COOP_GIF_URL), probability=0.3),
    Handler(Triggers.text.regexp(palti_you_are_regex), Responses.match.message('אתה {0}'), probability=0.4),
    Handler(Triggers.filter.user(users['palti']) & Triggers.text.regexp('^(חח)$'), Responses.preset.reply('בטח אתה כותב לי שני ח\' יא זונה קטנה')),
    Handler(Triggers.filter.user(users['ravid']) & Triggers.filter.video_note, Responses.preset.message('#רביד_מסריטה')),
    Handler(Triggers.event_streak(counting_event_trigger=Triggers.filter.user(users['palti']),
                                  resetting_event_trigger=None,
                                  streak_timeout_sec=60,
                                  event_count=5),
            Responses.preset.message('יואווו פלטי סתום כבר סתוםםם')),
]



import re
from ..dict_enum import DictEnum
from typing import List, Union
from telegram import Update
from .basic_triggers import BaseTrigger
from ..trigger_result import TriggerResult

class _RegExpTrigger(BaseTrigger):
    def __init__(self, pattern: str):
        self.pattern = pattern

    def check_trigger(self, update: Update) -> TriggerResult:
        if update.message.text is not None:
            text = update.message.text.replace('\n', ' ')
            match = re.match(self.pattern, text)
            if match is not None:
                return TriggerResult(should_respond=True,
                                     response_payload={ 'match': match.groups() })

        return TriggerResult(should_respond=False)


class _HasSubstringTrigger(_RegExpTrigger):
    def __init__(self, substring: Union[str, List[str]], exact: bool = False):
        if isinstance(substring, str):
            regexp_for_substring = substring
        elif isinstance(substring, list):
            if self.validate_strings_list(substring):
                regexp_for_substring = '|'.join(substring)
            else:
                raise TypeError('Parameter \'substring\' should be either a string or a list of strings. '
                                'A list containing non-strings was given)')
        else:
            raise TypeError(f'Parameter \'substring\' should be either a string or a list of strings '
                            f'({type(substring)} given)')

        prefix = '(?:.* )?' if exact else '.*'
        postfix = '(?: .*)?' if exact else '.*'
        super().__init__(f'^{prefix}({regexp_for_substring}){postfix}$')

    @staticmethod
    def validate_strings_list(strings):
        return all(map(lambda element: isinstance(element, str), strings))


class _HasExactWordTrigger(_HasSubstringTrigger):
    def __init__(self, word: Union[str, List[str]]):
        try:
            super().__init__(word, exact=True)
        except TypeError:
            raise TypeError(f'Parameter \'word\' should be either a string or a list of strings '
                            f'({type(word)} given)')


class TextTriggers(DictEnum):
    regexp = _RegExpTrigger
    has_substring = _HasSubstringTrigger
    has_exact_word = _HasExactWordTrigger



from ..dict_enum import DictEnum
from .preset_responses import PresetResponses
from .match_responses import MatchResponses

class Responses(DictEnum):
    preset = PresetResponses
    match = MatchResponses



import random
from ..dict_enum import DictEnum
from typing import Union, List
from telegram import Bot, Update, Document
from .basic_responses import BaseResponse
from .response_helper import ResponseHelper


class _PresetTextResponse(BaseResponse):
    def __init__(self, preset_response: Union[str, List[str]]):
        self.preset_responses = preset_response

    def get_response_text(self):
        if isinstance(self.preset_responses, str):
            return self.preset_responses
        elif isinstance(self.preset_responses, list):
            return random.choice(self.preset_responses)


class _PresetReplyResponse(_PresetTextResponse):
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.reply(bot, update, self.get_response_text())


class _PresetMessageResponse(_PresetTextResponse):
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.message(bot, update, self.get_response_text())


class _PresetDocumentResponse(BaseResponse):
    def __init__(self, preset_response: Union[str, Document]):
        self.preset_response = preset_response

    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.document(bot, update, self.preset_response)


class PresetResponses(DictEnum):
    reply = _PresetReplyResponse
    message = _PresetMessageResponse
    document = _PresetDocumentResponse



import json
from .configuration import users_file_path


class UsersHelper:

    def __init__(self, file_path: str = None):
        if file_path is not None:
            self.file_path = file_path
        else:
            self.file_path = users_file_path()

        self._users = {}
        self.load_users()

    def load_users(self, force: bool = False):
        if force or len(self._users) == 0:
            with open(self.file_path, 'r') as f:
                self._users = json.load(f)

    def get_user_id_by_nickname(self, nickname: str):
        return self._users[nickname]


default_users_helper = UsersHelper()


import abc
import typing
from inspect import isclass
from .boolean_helper import BooleanHelper
from ..triggers.basic_triggers import BaseTrigger
from ..responses import Responses
from ..triggers import Triggers


class BaseParser(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def mapping_class():
        """ Returns the mapping class (`Triggers` or `Responses`)"""
        pass

    @classmethod
    def parse_single(cls, config, globals):
        config_copy = dict(config)
        if 'name' in config_copy:
            config_copy.pop('name')

        mapping_cls = cls.mapping_class()
        element_cls = mapping_cls[config_copy.pop('type')]

        return element_cls(**config_copy)

    @classmethod
    def parse_many(cls, config, globals):
        return {
            element['name']: cls.parse_single(element, globals)
            for element
            in config
        }


class TriggerParser(BaseParser):
    @staticmethod
    def mapping_class():
        return Triggers

    @classmethod
    def parameters_to_parse_as_trigger(cls, config):
        trigger_class = Triggers[config['type']]
        parameters_type_hints = typing.get_type_hints(trigger_class.__init__)
        parameters_to_parse = []

        for parameter_name in config:

            # Meta-parameters to ignore
            if parameter_name in ['type', 'name']:
                continue

            # If this parameter has not type hint, we cannot check if it's a trigger or not
            if parameter_name not in parameters_type_hints:
                continue

            type_hint = parameters_type_hints[parameter_name]

            origin_type = getattr(type_hint, '__origin__', None)

            if origin_type is not None:
                if origin_type == typing.Union:
                    if BaseTrigger in type_hint.__args__:
                        parameters_to_parse.append(parameter_name)
                else:
                    raise NotImplementedError(f'Origin type {origin_type} is currently not supported')
            elif isclass(type_hint) and issubclass(type_hint, BaseTrigger):
                parameters_to_parse.append(parameter_name)

        return parameters_to_parse


    @classmethod
    def parse_single(cls, config, globals):
        return BooleanHelper.parse_subrule_as_trigger_or_response(config,
                                                                  globals,
                                                                  cls.parse_single_recursively)
    @classmethod
    def parse_single_recursively(cls, config, globals):
        if isinstance(config, str):
            return globals[config]

        config_copy = dict(config)
        parameters_to_parse = cls.parameters_to_parse_as_trigger(config)
        config.pop('type')

        for parameter in parameters_to_parse:
            config_copy[parameter] = cls.parse_single(config[parameter], globals)

        return super().parse_single(config_copy, globals)


class ResponseParser(BaseParser):
    @staticmethod
    def mapping_class():
        return Responses



from ..dict_enum import DictEnum
from .filter_triggers import FilterTriggers
from .text_triggers import TextTriggers
from .event_streak_trigger import EventStreakTrigger


class Triggers(DictEnum):
    text = TextTriggers
    filter = FilterTriggers
    event_streak = EventStreakTrigger



from enum import Enum, EnumMeta


class DictEnumMeta(EnumMeta):
    def __getitem__(self, key):
        key_path = key.split('.')
        item = super().__getitem__(key_path.pop(0)).value

        if len(key_path) > 0:
            return item['.'.join(key_path)]
        else:
            return item


class DictEnum(Enum, metaclass=DictEnumMeta):
    pass



from dataclasses import dataclass
from typing import Dict, Type
from .trigger_response import TriggerResponse
from .triggers_reponses_parsers import BaseParser, TriggerParser, ResponseParser
from ..triggers.basic_triggers import BaseTrigger
from ..responses.basic_responses import BaseResponse


@dataclass
class TriggerResponseParams:
    singular_key: str
    globals: Dict[str, TriggerResponse]
    parser: Type[BaseParser]

    @property
    def plural_key(self):
        return self.singular_key + 's'


@dataclass
class TriggerParams(TriggerResponseParams):
    def __init__(self, globals: Dict[str, BaseTrigger]):
        super().__init__(singular_key='trigger', globals=globals, parser=TriggerParser)


@dataclass
class ResponseParams(TriggerResponseParams):
    def __init__(self, globals: Dict[str, BaseResponse]):
        super().__init__(singular_key='response', globals=globals, parser=ResponseParser)



from typing import Dict, Union, Callable
from boolean import boolean
from ruamel_yaml.comments import CommentedMap
from .boolean_operators import OPERATOR_TYPE_TO_FUNCTION
from .trigger_response import TriggerResponse

class BooleanHelper:

    @staticmethod
    def evaluate_boolean_expression(expr: boolean.Expression,
                                    globals: Dict[str, TriggerResponse]) -> TriggerResponse:
        # If the trigger/response here is just a name, look for it in the globals
        if isinstance(expr, boolean.Symbol):
            return globals[str(expr)]

        boolean_function = OPERATOR_TYPE_TO_FUNCTION[type(expr)]
        evaluated_args = [BooleanHelper.evaluate_boolean_expression(arg, globals)
                          for arg
                          in expr.args]
        return boolean_function(*evaluated_args)

    @staticmethod
    def parse_subrule_as_trigger_or_response(subrule: Union[CommentedMap, str],
                                             globals: Dict[str, TriggerResponse],
                                             parsing_func: Callable[[Union[CommentedMap, str], Dict[str, TriggerResponse]], TriggerResponse]) -> TriggerResponse:
        if isinstance(subrule, str):
            algebra = boolean.BooleanAlgebra()
            parsed_expr = algebra.parse(subrule)
            return BooleanHelper.evaluate_boolean_expression(parsed_expr, globals)

        return parsing_func(subrule, globals)



from ..dict_enum import DictEnum
from telegram import Bot, Update
from .basic_responses import BaseResponse
from .response_helper import ResponseHelper


class _MatchTextResponse(BaseResponse):
    def __init__(self, template: str):
        self.template = template

    def build_response_text(self, response_payload: dict):
            return self.template.format(*response_payload['match'])


class _MatchReplyResponse(_MatchTextResponse):
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.reply(bot, update, self.build_response_text(response_payload))


class _MatchMessageResponse(_MatchTextResponse):
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> None:
        ResponseHelper.message(bot, update, self.build_response_text(response_payload))


class MatchResponses(DictEnum):
    reply = _MatchReplyResponse
    message = _MatchMessageResponse
